/*
 * Copyright (C) 2015 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <string.h>                     /* memmove() */
#include <sys/types.h>                  /* off_t */

#include "allocation.h"
#include "memutil.h"


#define MINIMUM_SIZE 256


static size_t get_alloc_size(Allocation* a,size_t request_size){
    size_t alloc_size;
    if(a->greedy){
        for(alloc_size=MINIMUM_SIZE;alloc_size<request_size;alloc_size<<=1);
    }else{
        alloc_size = request_size;
    }
    return alloc_size;
}


Allocation* allocation_init(Allocation* a,int greedy){
    a->size = 0;
    a->greedy = greedy;
    return a;
}


void allocation_done(Allocation* a){
    if(a->size)
        u_free(a->mem);
}



int allocation_try_ensure(Allocation* a,size_t request_size){
#ifdef USE_JEMALLOC
    size_t alloc_size;
    size_t new_size;
    if(request_size<=a->size)
        return a->size;
    alloc_size = get_alloc_size(a,request_size);
    new_size = xallocx(a->mem,request_size,alloc_size-request_size,0);
    if(new_size>=request_size)
        return a->size = new_size;
#endif
    return 0;
}


char* allocation_ensure(Allocation* a,size_t request_size,int preserve){
    if(request_size<=a->size)
        return a->mem;
    /* Special case if the allocation is empty, assume presereve is irrelevant */
    if(a->size<=0){
        a->size = get_alloc_size(a,request_size);
        if(a->size<=0)
            return NULL;
        return a->mem = u_new(char,a->size);
    }
#ifdef USE_JEMALLOC
    /* allocation_try_ensure() might settle for a smaller but big enough size */
    if(allocation_try_ensure(a,request_size))
        return a->mem;
#endif
    a->size = get_alloc_size(a,request_size);
    if(preserve){
        a->mem = u_renew(char,a->mem,a->size);
    }else{
        u_free(a->mem);
        a->mem = u_new(char,a->size);
    }
    return a->mem;
}


char* allocation_ensure_full(Allocation* a,size_t request_size,const char* preserve_begin,size_t preserve_size){
    off_t offs;
    /* Special case if the allocation is empty, assume preserve_begin and
       preserve_size cannot be valid. */
    if(a->size<=0){
#ifdef USE_JEMALLOC
        a->mem = (char*)mallocx(get_alloc_size(a,request_size),0);
        a->size = sallocx(a->mem,0);
#else
        a->size = get_alloc_size(a,request_size);
        return a->mem = u_new(char,a->size);
#endif
        return a->mem;
    }
    if(!preserve_begin)
        preserve_begin = a->mem;
    offs = preserve_begin - a->mem;
    if(offs+request_size<=a->size)
        return (char*)preserve_begin;
#ifdef USE_JEMALLOC
    if(allocation_try_ensure(a,offs+request_size)){
        /* The allocation could be expanded in place. */
        return (char*)preserve_begin;
    }else if(request_size<=a->size){
        /* The current allocation could be used by moving the used part. */
        memmove(a->mem,preserve_begin,preserve_size);
        return a->mem;
    }else{
        /* Make a new allocation. */
        char* tmp = a->mem;
        a->mem = (char*)mallocx(get_alloc_size(a,request_size),0);
        a->size = sallocx(a->mem,0);
        if(preserve_size>0)
            memcpy(a->mem,preserve_begin,preserve_size);
        u_free(tmp);
        return a->mem;
    }
#else
    if(request_size<=a->size){
        /* The current allocation could be used by moving the used part. */
        memmove(a->mem,preserve_begin,preserve_size);
        return a->mem;
    }else{
        /* Make a new allocation. */
        a->size = get_alloc_size(a,offs+request_size);
        a->mem = u_renew(char,a->mem,a->size);
        return a->mem+offs;
    }
#endif
}
