/*
 * Copyright (C) 2015 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
//#define TEST_EVENTFD
#undef TEST_EVENTFD

#include "debug.h"
#include "eventloop.h"
#include "list.h"
#include "memutil.h"
#include "queue.h"

#ifdef EVENTLOOP_USE_GLIB

#else

#include <errno.h>
#ifdef TEST_EVENTFD
# include <poll.h>
# include <sys/eventfd.h>
# include <unistd.h>
#endif

#include "asyncfifo.h"

struct _Eventloop{
#ifdef TEST_EVENTFD
    int wakeup_fd;
#endif
    AsyncFifo* dispatch_fifo;
    int quit;
    List* fd_list;
    Queue* timer_queue;
};


typedef struct{
    DispatchFunc func;
    void* data;
}DispatchData;


#ifdef TEST_EVENTFD
static int on_wakeup(void* userdata,int fd){
    uint64_t value;
    (void)(read(fd,&value,sizeof(uint64_t))<0);
    return 1;
}


static void wakeup(Eventloop* loop){
    uint64_t value = 1;
    /* force ignore return value */
    (void)(write(loop->wakeup_fd,&value,sizeof(uint64_t))+1);
}
#endif


static int eventloop_new_cb(Eventloop* loop){
    DispatchData* dd = (DispatchData*)async_fifo_pop(loop->dispatch_fifo);
    if(dd){
        dd->func(dd->data);
        u_free(dd);
    }else{
        loop->quit = 1;
    }
    return 1;
}
Eventloop* eventloop_new(){
    Eventloop* loop = u_new(Eventloop,1);
#ifdef TEST_EVENTFD
    loop->wakeup_fd = eventfd(0,EFD_NONBLOCK|EFD_SEMAPHORE);
    eventloop_add_fd(loop,loop->wakeup_fd,on_wakeup,NULL);
#endif
    loop->dispatch_fifo = async_fifo_new();
    loop->fd_list = NULL;
    loop->timer_queue = queue_new();
    eventloop_add_fd(loop,async_fifo_fd(loop->dispatch_fifo),(EventloopFdFunc)eventloop_new_cb,loop);
    return loop;
}


void eventloop_free(Eventloop* loop){
#ifdef TEST_EVENTFD
    close(loop->wakeup_fd);
#endif
    async_fifo_free(loop->dispatch_fifo);
    queue_free(loop->timer_queue);
    u_free(loop);
}

#endif /* EVENTLOOP_USE_GLIB */


typedef struct{
    int fd;
    EventloopFdFunc func;
    void* userdata;
}EventloopAddFdData;
#if 0
  /* FIXME!! */
#ifdef EVENTLOOP_USE_GLIB
static gboolean eventloop_add_fd_cb(int fd,GIOCondition condition,void* userdata)
#else
static int eventloop_add_fd_cb(int fd,void* userdata)
#endif
{
    EventloopAddFdData* data = (EventloopAddFdData*)userdata;
    if(data->func(data->userdata,fd)){
        u_free(data);
        return FALSE;
    }else{
        return TRUE;
    }
}
#endif
#ifndef TEST_EVENTFD
static void eventloop_add_fd_cb(void* data){
    /* do nothing */
}
#endif
void eventloop_add_fd(Eventloop* loop,int fd,EventloopFdFunc func,void* userdata){
    EventloopAddFdData* data = u_new(EventloopAddFdData,1);
    data->func = func;
    data->userdata = userdata;
#ifdef EVENTLOOP_USE_GLIB
    /* FIXME!! */
#else
    data->fd = fd;
    loop->fd_list = list_prepend(loop->fd_list,data);
# ifdef TEST_EVENTFD
    wakeup(loop);
# else
    eventloop_dispatch(loop,eventloop_add_fd_cb,data);
# endif
#endif
}


void eventloop_add_timeout(Eventloop* loop,int timout,EventloopTimeoutFunc func,void* data){
    /* FIXME!! */
}


int eventloop_run(Eventloop* loop){
#ifdef EVENTLOOP_USE_GLIB
    g_main_loop_run(loop);
#else
    loop->quit = 0;
    while(!loop->quit){
        fd_set fdread;
        List* iter;
        int maxfd;
        int res;

        FD_ZERO(&fdread);

        maxfd = -1;
        for(iter=loop->fd_list;iter;iter=iter->next){
            EventloopAddFdData* event = (EventloopAddFdData*)iter->data;
            if(event->fd>maxfd)
                maxfd = event->fd;
            FD_SET(event->fd,&fdread);
        }

        /* FIXME!! add timeout... */

        res = select(maxfd+1,&fdread,NULL,NULL,NULL);
        if(res<0){
            if(EINTR==errno)
                continue;
            DEBUG_p("select failed: %s",strerror(errno));
            break;
        }

        for(iter=loop->fd_list;res>0 && iter;){
            EventloopAddFdData* event = (EventloopAddFdData*)iter->data;
            List* next = iter->next;
            if(FD_ISSET(event->fd,&fdread)){
                if(!event->func(event->userdata,event->fd)){
                    loop->fd_list = list_remove(loop->fd_list,iter);
                    u_free(event);
                }
                --res;
            }
            iter = next;
        }

        /* FIXME!! check timeout events */
    }
#endif
    return 0;
}


void eventloop_quit(Eventloop* loop){
#ifdef EVENTLOOP_USE_GLIB
    g_main_loop_quit(loop);
#else
    /* FIXME!! check if running in same thread */
#if 1
    async_fifo_push(loop->dispatch_fifo,NULL);
#else
    loop->quit = 1;
#endif
#endif
}



void eventloop_dispatch(Eventloop* loop,DispatchFunc func,void* data){
#ifdef EVENTLOOP_USE_GLIB
    /* FIXME!! */
#else
    DispatchData* dd = u_new(DispatchData,1);
    dd->func = func;
    dd->data = data;
    async_fifo_push(loop->dispatch_fifo,dd);
#endif
}
