/*
 * Copyright (C) 2016 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
/** @file

  Code for generating JavaScrupt code.

  In JS each widget is represented by an object which may correspond to an
  element in the DOM. Each object has the following attributes:

    mw: Minimum width, excluding the padding (and border) around the object.
    mh: Minimum height, excluding the padding.
    pw: Padding width (left + right). (optional)
    ph: Padding height (top + bottom). (optional)
    ew: Expand width, non zero value and the element will be horisontally
        expanded. (optional)
    eh: Expand height. (optional)
    fw: Stretch the widget (content) to the allocated width. (optional)
    fh: Stretch the widget (content) to the allocated height. (optional)
    onr: Custom geometry request function. (optional)
    ons: Cutsom geometry set function. (optional)
    cx: Cell X position in table. (only then in a table)
    cy: Cell Y position in table. (only then in a table)
    cw: Number of cells wide; (only then in a table)
    ch: Number of cells high; (only then in a table)

  During runtime the following attributes are also used:

   (x:  Current (calculated) x position of the element.)
   (y:  Current (calculated) y position of the element.)
    w:  Total width of the element.
    h:  Total height of the element.
    ew: Calculated if not set.
    eh: Calculated if not set.

  After the request geometry function is called the attributes w, mh, h
  and eh must be set on the element. A custom geometry request function is
  responsibe for setting them.

*/
#include "config.h"

#include <stdio.h>

#include "debug.h"
#include "genjs.h"
#include "hash.h"
#include "memutil.h"
#include "session.h"

/* --- functions -------------------------------------------------------- */

const char* genjs_get_function(Session* sess,GenjsFunction func){
    const char* name = session_get_jsid(sess,func);
    if(!name){
        name = session_add_jsid(sess,func);
        func(sess,name);
    }
    return name;
}




#if 0
static void func_model_set(WguiRequest* r,const char* id){
    fprintf(r->fd,"function %s(m,val){",id);
    fprintf(r->fd,  "if(m.value != val){"
                      "m.value = value;"
                      "for(var i=0;i<m.hooks.length;++i)"
                        "m.hooks[i].widget[m.hooks[i].property] = val;"
                    "}"
                  "}");
}
#endif

/* --- data models ------------------------------------------------------ */

typedef struct{
    void* model;
    Session* session;
}ModelSession;

/* --- list model ------------------------------------------------------- */

static void create_list_model_add(Session* sess,const char* list_model_add){
    out("%s=function(m,i,v){",list_model_add);
    out(  "if(i<0||i>=m.item.length){"
            "i=m.item.length;"
            "m.item[i]=v;"
          "}else{"
            "m.item.splice(i,0,v);"
          "}"
          "for(var j=0;j<m.on_add.length;++j)"
            "m.on_add[j](i,v);"
        "}");
}
#define get_list_model_add() genjs_get_function(sess,create_list_model_add)


static void create_list_model_del(Session* sess,const char* list_model_del){
    out("%s=function(m,i){",list_model_del);
    out(  "if(i<0||i>m.item.length)"
            "return;"
          "m.item.splice(i,1);"
          "for(var j=0;j<m.on_del.length;++j)"
            "m.on_add[j](i);"
        "}");
}
#define get_list_model_del() genjs_get_function(sess,create_list_model_del)


/** Called from js. */
static void up_list_model_on_add(void* userdata,Session* sess,const char* args){
    DEBUG_p("args: %s",args);
}


/** Called from js. */
static void up_list_model_on_del(void* userdata,Session* sess,const char* args){
    DEBUG_p("args: %s",args);
}


/** Calls to js. */
static int on_list_model_add(void* userdata,void* data,int num){
    ModelSession* ms = (ModelSession*)userdata;
    Session* sess = ms->session;
    out("%s(%s,%d,'%s');",get_list_model_add(),genjs_get_list_model(sess,ms->model),num,data);
    return 0;
}


/** Calls to js. */
static int on_list_model_del(void* userdata,void* data,int num){
    ModelSession* ms = (ModelSession*)userdata;
    Session* sess = ms->session;
    out("%s(%s,%d);",get_list_model_del(),genjs_get_list_model(sess,ms->model),num);
    return 0;
}


const char* genjs_get_list_model(Session* sess,ListModel* m){
    const char* name = session_get_jsid(sess,m);
    if(!name){
        ModelSession* ms = u_new(ModelSession,1);
        /* FIXME!! add some pointer to ms in the session so it can be freed */
        name = session_add_jsid(sess,m);
        out("%s={item:[],on_add:[function(i,v){_('%s#'+i+'$'+v);}],on_del:[function(i){_('<%s'+i);}]};",
            name,
            session_register_js_callback(sess,up_list_model_on_add,m),
            session_register_js_callback(sess,up_list_model_on_del,m));
        ms->model = m;
        ms->session = sess;
        list_model_subscribe_add(m,on_list_model_add,ms);
        list_model_subscribe_remove(m,on_list_model_del,ms);
    }
    return name;
}

/* --- range model ------------------------------------------------------ */

void genjs_create_range_model_set(Session* sess,const char* range_model_set){
}


void genjs_create_range_model_limits(Session* sess,const char* range_model_limits){
}


/** Called from js. */
static void up_range_model_on_changed(void* userdata,Session* sess,const char* args){
}


/** Called from js. */
static void up_range_model_on_limits(void* userdata,Session* sess,const char* args){
}


/** Calls to js. */
static int on_range_model_changed(void* userdata){
    /* FIXME!! */
    return 0;
}


/** Calls to js. */
static int on_range_model_limits(void* userdata){
    /* FIXME!! */
    return 0;
}


const char* genjs_get_range_model(Session* sess,RangeModel* m){
    const char* name = session_get_jsid(sess,m);
    if(!name){
        ModelSession* ms = u_new(ModelSession,1);
        /* FIXME!! add some pointer to ms in the session so it can be freed */
        name = session_add_jsid(sess,m);
        out("%s={min:%d,max:%d,v:%d,ons:[function(){_('%s')}],onl:[function(){_('%s')}]};",
            name,
            m->min,
            m->max,
            m->value,
            session_register_js_callback(sess,up_range_model_on_changed,m),
            session_register_js_callback(sess,up_range_model_on_limits,m));
        ms->model = m;
        ms->session = sess;
        range_model_subscribe_changed(m,on_range_model_changed,ms);
        range_model_subscribe_limits(m,on_range_model_limits,ms);
    }
    return name;
}

/* --- element ---------------------------------------------------------- */

/**
  Try to lookup a widget of unknown type, which has to be preexisting.
*/
const char* genjs_get_element(UpWidget* w){
    Session* sess = w->session;
    const char* name = session_get_jsid(sess,w);
    if(!name){
        name = session_add_jsid(sess,w);
        w->vt->genjs(w,name);
    }
    return name;
}

/* --- main functions --------------------------------------------------- */

void genjs_init(Session* sess){
    const char* element_set_geometry;
    out("document.documentElement.style.overflow='hidden';"
        "document.body.scroll='no';");
    element_set_geometry = genjs_get_element_set_geometry();
    out("window.onresize=function(){%s(document.body,0,0,this.innerWidth,this.innerHeight);};",element_set_geometry);
}


void genjs_comet(Session* sess){
#ifdef TEST_COMET
    /* open longread connection */
    session_send(sess,"_('+');",-1);
#endif
}


void genjs_resize(UpWidget* w){
    Session* sess = w->session;
    const char* element_set_geometry = genjs_get_element_set_geometry();
    out("%s(%s,0,0,this.innerWidth,this.innerHeight);",element_set_geometry,genjs_get_element(w));
}
