/*
 * Copyright (C) 2016 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#ifndef USE_MICROHTTPD

#include "priv.h"

#include <fcntl.h>                      /* open() */
#include <netinet/in.h>
#include <sys/sendfile.h>               /* sendfile() */
#include <sys/socket.h>
#include <sys/stat.h>                   /* fstat() */
//#include <sys/un.h>

#include "debug.h"
#include "fileinfo.h"
#include "list.h"
#include "mainloop.h"
#include "httpd.h"
#include "strn.h"

#define BUFFER_SIZE 512


struct _Httpd{
    int server_socket;
    HttpdRequestHandler handler;
    void* userdata;
};

struct _HttpdConnection{
    Httpd* httpd;
    Strn buffer;                        /**< Buffer for the request. */
    int fd;
    char* method;
    char* url;
    char* args;
    char* version;
    List* cookie_list;
};

typedef struct{
    const char* name;
    int name_len;
    const char* value;
    int value_len;
}Cookie;


static HttpdConnection* httpd_connection_new(Httpd* httpd,int fd){
    HttpdConnection* c = u_new(HttpdConnection,1);
    c->httpd = httpd;
    strn_init(&c->buffer,1);
    c->fd = fd;
    c->method = NULL;
    c->url = NULL;
    c->args = NULL;
    c->version = NULL;
    c->cookie_list = NULL;
    return c;
}


static void httpd_connection_free(HttpdConnection* c){
    strn_done(&c->buffer);
    close(c->fd);
    u_free_if_set(c->method);
    u_free_if_set(c->url);
    u_free_if_set(c->args);
    u_free_if_set(c->version);
    u_free(c);
}


static void add_cookies(HttpdConnection* c,const char* s,int len){
    const char* name;
    int name_len;
    const char* value;
    int value_len;
    const char* p = s;
    const char* end = p+len;
    while(p<end){
        Cookie* cookie;
        for(;' '==*p||';'==*p;++p);
        if(p>=end)
            break;
        name = p;
        for(;*p!='=';++p);
        name_len = p-name;
        ++p;
        value = p;
        for(;*p!=';'&&p<end;++p);
        value_len = p-value;
        DEBUG_p("'%.*s': '%.*s'",name_len,name,value_len,value);
        cookie = u_new(Cookie,1);
        cookie->name = name;
        cookie->name_len = name_len;
        cookie->value = value;
        cookie->value_len = value_len;
        c->cookie_list = list_prepend(c->cookie_list,cookie);
    }
}

static void parse_request(HttpdConnection* c){
    const char* name;
    int name_len;
    const char* value;
    int value_len;
    const char* p = c->buffer.str;
    const char* end = p+c->buffer.len;
    value = p;
    for(;*p!=' ';++p);
    c->method = u_strndup(value,p-value);
    for(;' '==*p;++p);
    value = p;
    for(;*p!=' '&&*p!='?';++p);
    c->url = u_strndup(value,p-value);
    if('?'==*p){
        value = ++p;
        for(;*p!=' ';++p);
        c->args = u_strndup(value,p-value);
    }
    for(;' '==*p;++p);
    value = p;
    for(;*p!='\r'&&*p!='\n';++p);
    c->version = u_strndup(value,p-value);
    while(p<end){
        for(;'\r'==*p||'\n'==*p;++p);
        if(p>=end)
            break;
        name = p;
        for(;*p!=':';++p);
        name_len = p-name;
        for(;':'==*p||' '==*p;++p);
        value = p;
        for(;*p!='\r'&&*p!='\n';++p);
        value_len = p-value;
        DEBUG_p("'%.*s': '%.*s'",name_len,name,value_len,value);
        if(!strncmp("Cookie",name,name_len)){
            add_cookies(c,value,value_len);
        }
    }
}



static int on_data(void* userdata,int fd){
    HttpdConnection* c = (HttpdConnection*)userdata;
    ssize_t len;
    strn_ensure_writable_length(&c->buffer,c->buffer.len+BUFFER_SIZE,1);
    len = read(fd,c->buffer.str+c->buffer.len,BUFFER_SIZE);
    if(len<0){
        /* error */
        return 0;
    }
    /* successful read */
    c->buffer.len += len;
    if(len<BUFFER_SIZE){
        /* end of file */
        parse_request(c);
        c->httpd->handler(c->httpd->userdata,c,c->url,c->method,c->version);
        return 0;
    }
    return 1;
}


static int on_connect(void* userdata,int fd){
    int client_socket;
    HttpdConnection* c;
    DEBUG_p("begin");
    client_socket = accept(fd,NULL,0);
    if(client_socket<0)
        return 0;
    fcntl(client_socket,F_SETFD,FD_CLOEXEC);
    c = httpd_connection_new((Httpd*)userdata,client_socket);
    mainloop_add_fd(c->fd,on_data,c);
    return 1;
}


Httpd* httpd_new(int port,HttpdRequestHandler handler,void* userdata){
    int server_socket;
    static const int const1 = 1;
    struct sockaddr_in server_address;
    Httpd* httpd;

    /* do network stuff */
    server_socket = socket(AF_INET,SOCK_STREAM,0);
    if(server_socket<0){
        DEBUG_p("socket() failed");
        return NULL;
    }
    setsockopt(server_socket,SOL_SOCKET,SO_REUSEADDR,&const1,sizeof(int));
    memset(&server_address,0,sizeof(struct sockaddr_in));
    server_address.sin_family = AF_INET;
    server_address.sin_port = htons(port);
    server_address.sin_addr.s_addr = htons(INADDR_ANY);
    if(bind(server_socket,(struct sockaddr*)&server_address,sizeof(struct sockaddr_in))<0){
        DEBUG_p("bind() failed");
        close(server_socket);
        return NULL;
    }
    listen(server_socket,20);
    httpd = u_new(Httpd,1);
    httpd->server_socket = server_socket;
    httpd->handler = handler;
    httpd->userdata = userdata;
    mainloop_add_fd(server_socket,on_connect,httpd);
    return httpd;
}


void httpd_free(Httpd* httpd){
    close(httpd->server_socket);
    u_free(httpd);
}



const char* httpd_get_args(HttpdConnection* c){
    return c->args;
}


char* httpd_get_cookie(HttpdConnection* c,const char* name){
    List* iter;
    for(iter=c->cookie_list;iter;iter=iter->next){
        Cookie* cookie = (Cookie*)iter->data;
        if(!strncmp(name,cookie->name,cookie->name_len))
            return u_strndup(cookie->value,cookie->value_len);
    }
    return NULL;
}



static void httpd_header_begin(HttpdConnection* c,const char* status){
    dprintf(c->fd,"HTTP/1.0 %s\r\n",status);
    dprintf(c->fd,"Server: " PACKAGE_APPNAME "/" PACKAGE_VERSION "\r\n");
}


void httpd_header_ok(HttpdConnection* c){
    httpd_header_begin(c,"200 OK");
}


void httpd_header_not_found(HttpdConnection* c){
    httpd_header_begin(c,"404 NOT FOUND");
}


void httpd_header(HttpdConnection* c,const char* name,const char* value){
    dprintf(c->fd,"%s: %s\r\n",name,value);
}


void httpd_header_cookie(HttpdConnection* c,const char* name,const char* value){
    dprintf(c->fd,"Set-Cookie: %s=%s\r\n",name,value);
}


void httpd_header_end(HttpdConnection *c){
    (void)(write(c->fd,"\r\n",2)<0);
}


/*
int httpd_steal_fd(HttpdConnection* c){
    int fd = req->fd;
    req->fd = -1;
    return fd;
}
*/


void httpd_connection_close(HttpdConnection* c){
    httpd_connection_free(c);
}



int httpd_send(HttpdConnection* c,const char* data,int len){
    if(len<0)
        len = strlen(data);
    return write(c->fd,data,len);
}


int httpd_sendf(HttpdConnection* c,const char* format,...){
    va_list ap;
    int len;
    va_start(ap,format);
    len = vdprintf(c->fd,format,ap);
    va_end(ap);
    return len;
}


int httpd_sendvf(HttpdConnection* c,const char* format,va_list ap){
    return vdprintf(c->fd,format,ap);
}


int httpd_sendfile(HttpdConnection* c,const char* filename,const char* mimetype){
    struct stat buf;
    int fd = open(filename,O_RDONLY);
    if(fd<0)
        goto notfound;
    if(fstat(fd,&buf)<0){
        close(fd);
        goto notfound;
    }
    httpd_header_ok(c);
    dprintf(c->fd,"Content-Length: %d\r\n",(int)buf.st_size);
    if(mimetype){
        httpd_header(c,"Content-Type",mimetype);
    }else{
        char* mimetype2;
        get_fileinfo(filename,&mimetype2);
        if(mimetype2){
            httpd_header(c,"Content-Type",mimetype);
            u_free(mimetype2);
        }
    }
    httpd_header_end(c);
    sendfile(c->fd,fd,0,buf.st_size);
    return buf.st_size;
notfound:
    httpd_header_not_found(c);
    httpd_header_end(c);
    return -1;
}

#endif /* ! USE_MICROHTTPD */
