/*
 * Copyright (C) 2016 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdlib.h>                     /* qsort_r() */

#include "array.h"
#include "listmodel.h"
#include "memutil.h"
#include "queue.h"
#include "uref.h"

#define TEST_FILTER_ARRAY


typedef struct{
    void* data;
    int num;
}ModelItemData;
static int model_signal_invoke_cb(SignalHandler handler,void* userdata,void* signaldata){
    ModelItemData* d = (ModelItemData*)signaldata;
    ModelSignalHandler cb = (ModelSignalHandler)handler;
    return cb(userdata,d->data,d->num);
}
int model_signal_invoke(Signal* signal,void* data,int num){
    ModelItemData d;
    d.data = data;
    d.num = num;
    return signal_invoke_marshal(signal,model_signal_invoke_cb,&d);
}

/* --- ListModel -------------------------------------------------------- */

static void list_model_free_internal(ListModel* m){
    m->vt->done(m);
    u_free(m);
}


ListModel* list_model_init(ListModel* m){
    ((URef*)m)->count = 1;
    ((URef*)m)->destroy = (URefDestroyFunc)list_model_free_internal;
    signal_init(&m->signal_add);
    signal_init(&m->signal_remove);
    return m;
}


void list_model_done_func(ListModel* m){
    signal_done(&m->signal_add);
    signal_done(&m->signal_remove);
}


SignalSubscription* list_model_subscribe_add(ListModel* m,ModelSignalHandler cb,void* userdata){
    return signal_subscribe(&m->signal_add,(SignalHandler)cb,userdata);
}


SignalSubscription* list_model_subscribe_remove(ListModel* m,ModelSignalHandler cb,void* userdata){
    return signal_subscribe(&m->signal_remove,(SignalHandler)cb,userdata);
}

/* --- ListModelRow ----------------------------------------------------- */

#if 0
/**
  Utility function which takes advantage of cases where the row number is
  already known.
*/
static void list_model_row_set_with_num(ListModelRow* r,ListModelIter* iter,int num){
    if(r->iter==iter)
        return;
    if(iter){
        if(num<0)
            num = list_model_get_num(r->model,iter);
        r->data = list_model_get(r->model,iter);
        r->num = num;
    }
    r->iter = iter;
}


#ifndef TEST_FILTER_ARRAY
/**
  Utility function which takes advantage of cases where the row number is
  already known.
*/
static ListModelRow* list_model_row_new_with_num(ListModel* child,ListModelIter* iter,int num){
    ListModelRow* row = list_model_row_new(child);
    list_model_row_set_num(row,iter,num);
    return row;
}
#endif
#endif


#define list_model_row_validate_iter(r) if(!(r)->iter)(r)->iter=list_model_get_iter((r)->model,(r)->num)


static int list_model_row_on_add(void* userdata,void* data,int num){
    ListModelRow* r = (ListModelRow*)userdata;
    if(list_model_length(r->model)<=1){
        /* if no row was selected, select the first one */
        r->num = num;
    }else{
        if(num<=r->num)
            ++r->num;
    }
    /* invalidate iter since it may no longer be valid */
    r->iter = NULL;
    return 0;
}


static int list_model_row_on_remove(void* userdata,void* data,int num){
    ListModelRow* r = (ListModelRow*)userdata;
#if 1
    /* FIXME!! is r->num<0 really valid to expect in this case? */
    //if(r->num<0)
    //    return 0;
    if(list_model_length(r->model)<=1){
        /* this is the last item */
        r->num = -1;
    }else if(num<r->num){
        --r->num;
    }else if(r->num>=list_model_length(r->model)-1){
        --r->num;
    }
    /* invalidate iter since it may no longer be valid */
    r->iter = NULL;
#elif 1
    int comp = list_model_compare(r->model,iter,r->iter);
    if(0==comp){
        ListModelIter* iter2;
        /* this row was deleted */
        if((iter2=list_model_next(r->model,iter))){
            /* select the next row */
            list_model_row_set_num(r,iter2,r->num);
        }else if((iter2=list_model_previous(r->model,iter))){
            /* select the previous row */
            list_model_row_set_num(r,iter2,r->num-1);
        }else{
            /* no other row to select */
            r->iter = NULL;
            return 0;
        }
    }else if(comp<0){
        --r->num;
    }

#else
    if(num == r->num){
        ListModelIter* iter2;
        /* this row was deleted */
        if((iter2=list_model_next(r->model,iter))){
            /* select the next row */
            list_model_row_set_num(r,iter2,r->num);
        }else if((iter2=list_model_previous(r->model,iter))){
            /* select the previous row */
            list_model_row_set_num(r,iter2,r->num-1);
        }else{
            /* no other row to select */
            r->iter = NULL;
        }
    }else if(num < r->num){
        --r->num;
    }
#endif
    return 0;
}


ListModelRow* list_model_row_init(ListModelRow* r,ListModel* m){
    r->model = m;
    r->iter = NULL;
    if(m){
        if(list_model_length(m)<=0){
            /* empty model */
            r->num = -1;
        }else{
            /* select the first item */
            r->num = 0;
        }
        r->ss_add = list_model_subscribe_add(m,list_model_row_on_add,r);
        r->ss_remove = list_model_subscribe_remove(m,list_model_row_on_remove,r);
    }else{
        r->num = -1;
    }
    return r;
}


void list_model_row_done(ListModelRow* r){
    if(r->model){
        list_model_unsubscribe_add(r->model,r->ss_add);
        list_model_unsubscribe_remove(r->model,r->ss_remove);
        list_model_unref(r->model);
    }
}


ListModelRow* list_model_row_new(ListModel* m){
    return list_model_row_init(u_new(ListModelRow,1),m);
}


void list_model_row_free(ListModelRow* m){
    list_model_row_done(m);
    u_free(m);
}


void list_model_row_set(ListModelRow* r,ModelIter* iter){
    r->num = list_model_get_num(r->model,iter);
    r->iter = iter;
}


ModelIter* list_model_row_iter(ListModelRow* r){
    if(r->num<0)
        return NULL;
    if(!r->iter)
        r->iter = list_model_get_iter(r->model,r->num);
    return r->iter;
}


void* list_model_row_data(ListModelRow* r){
    if(list_model_row_valid(r)){
        return list_model_get(r->model,list_model_row_iter(r));
    }else{
        return NULL;
    }
}


void list_model_row_first(ListModelRow* r){
    r->num = 0;
    r->iter = list_model_first(r->model);
}


void list_model_row_last(ListModelRow* r){
    r->num = list_model_length(r->model)-1;
    r->iter = list_model_last(r->model);
}


int list_model_row_previous(ListModelRow* r){
    ModelIter* iter;
    list_model_row_validate_iter(r);
    iter = list_model_previous(r->model,r->iter);
    if(!iter)
        return 0;
    --r->num;
    r->iter = iter;
    return 1;
}


int list_model_row_next(ListModelRow* r){
    ModelIter* iter;
    list_model_row_validate_iter(r);
    iter = list_model_next(r->model,r->iter);
    if(!iter)
        return 0;
    ++r->num;
    r->iter = iter;
    return 1;
}

/* --- ListModelStore --------------------------------------------------- */

typedef struct{
    ListModel inh;

    Queue* item_queue;
    ModelItemNotifyFunc item_notify_func;

    Signal signal_changed;
}ListModelStore;


static void list_model_store_done(ListModel* m){
    ListModelStore* ms = (ListModelStore*)m;
    if(ms->item_notify_func){
        queue_free_full(ms->item_queue,ms->item_notify_func);
    }else{
        queue_free(ms->item_queue);
    }
    signal_done(&ms->signal_changed);
    list_model_done_func(m);
}


static int list_model_store_get_length(ListModel* m){
    ListModelStore* ms = (ListModelStore*)m;
    return queue_length(ms->item_queue);
}


static ModelIter* list_model_store_first(ListModel* m){
    ListModelStore* ms = (ListModelStore*)m;
    return (ModelIter*)ms->item_queue->head;
}


static ModelIter* list_model_store_last(ListModel* m){
    ListModelStore* ms = (ListModelStore*)m;
    return (ModelIter*)ms->item_queue->tail;
}


static ModelIter* list_model_store_previous(ListModel* m,ModelIter* iter){
    return (ModelIter*)((List*)iter)->prev;
}


static ModelIter* list_model_store_next(ListModel* m,ModelIter* iter){
    return (ModelIter*)((List*)iter)->next;
}


static int list_model_store_compare(ListModel* m,ModelIter* a,ModelIter* b){
    if(a==b)
        return 0;
    for(;;){
        a = list_model_store_previous(m,a);
        if(!a)
            return -1;
        if(a==b)
            return 1;
    }
}


static int list_model_store_foreach(ListModel* m,ModelForeachFunc callback,void* userdata){
    ListModelStore* ms = (ListModelStore*)m;
    int result;
    List* iter;
    int nr;
    for(iter=ms->item_queue->head,nr=0;iter;iter=iter->next,++nr){
        result = callback(userdata,iter->data,nr,(ModelIter*)iter);
        if(result)
            return result;
    }
    return 0;
}


ModelIter* list_model_store_get_iter(ListModel* m,int num){
    ListModelStore* ms = (ListModelStore*)m;
    List* item = ms->item_queue->head;
    while(num-->0){
        if(!item)
            return NULL;
        item = item->next;
    }
    return (ModelIter*)item;
}


int list_model_store_get_num(ListModel* m,ModelIter* iter){
    List* item = (List*)iter;
    int num = 0;
    while(item->prev){
        ++num;
        item = item->prev;
    }
    return num;
}


static void* list_model_store_get(ListModel* m,ModelIter* iter){
    return ((List*)iter)->data;
}


ListModel* list_model_store_new(ModelItemNotifyFunc func){
    static const ListModelInterface list_model_store_vt = {
        .done     = list_model_store_done,
        .length   = list_model_store_get_length,
        .first    = list_model_store_first,
        .last     = list_model_store_last,
        .previous = list_model_store_previous,
        .next     = list_model_store_next,
        .compare  = list_model_store_compare,
        .foreach  = list_model_store_foreach,
        .get_iter = list_model_store_get_iter,
        .get_num  = list_model_store_get_num,
        .get      = list_model_store_get
    };
    ListModelStore* ms = u_new(ListModelStore,1);
    ListModel* m = list_model_init((ListModel*)ms);
    DEBUG_ref_reg((URef*)ms,"ListModelStore");
    m->vt = &list_model_store_vt;

    ms->item_queue = queue_new();
    ms->item_notify_func = func;

    signal_init(&ms->signal_changed);
    return m;
}


void list_model_store_append(ListModel* m,void* data){
    ListModelStore* ms = (ListModelStore*)m;
    List* item = queue_push_tail(ms->item_queue,data);
    if(m->signal_add.subscription_list)
        model_signal_invoke(&m->signal_add,data,list_model_store_get_num(m,(ModelIter*)item));
}


void list_model_store_prepend(ListModel* m,void* data){
    ListModelStore* ms = (ListModelStore*)m;
    List* item = queue_push_head(ms->item_queue,data);
    if(m->signal_add.subscription_list)
        model_signal_invoke(&m->signal_add,data,list_model_store_get_num(m,(ModelIter*)item));
}


void list_model_store_insert_before(ListModel* m,ModelIter* iter,void* data){
    ListModelStore* ms = (ListModelStore*)m;
    List* item = queue_insert_before(ms->item_queue,(List*)iter,data);
    if(m->signal_add.subscription_list)
        model_signal_invoke(&m->signal_add,data,list_model_store_get_num(m,(ModelIter*)item));
}


void list_model_store_insert_after(ListModel* m,ModelIter* iter,void* data){
    ListModelStore* ms = (ListModelStore*)m;
    List* item = queue_insert_before(ms->item_queue,(List*)iter,data);
    if(m->signal_add.subscription_list)
        model_signal_invoke(&m->signal_add,data,list_model_store_get_num(m,(ModelIter*)item));
}


void* list_model_store_remove(ListModel* m,ModelIter* iter){
    ListModelStore* ms = (ListModelStore*)m;
    List* item = (List*)iter;
    void* itemdata = item->data;
    if(m->signal_remove.subscription_list)
        model_signal_invoke(&m->signal_remove,itemdata,list_model_store_get_num(m,iter));
    queue_remove(ms->item_queue,item);
    return itemdata;
}


void list_model_store_remove_data(ListModel* m,void* data){
    ListModelStore* ms = (ListModelStore*)m;
    List* iter;
    for(iter=ms->item_queue->head;iter;iter=iter->next){
        if(iter->data==data){
            list_model_store_remove(m,(ModelIter*)iter);
            return;
        }
    }
}


#if 0
/* --- ListModelSort ----------------------------------------------------- */

typedef struct{
    void* data;
    int num;
}ListModelSortNode;

struct _ListModelSort{
    ListModel inh;
    ListModel* child;
    Array* item_array;
    ModelCompareFunc compare_func;
    SignalSubscription* ss_add;
    SignalSubscription* ss_remove;
};


ListModelFilterNode* list_model_sort_node_new(ListModelSort* ms,int num,void* data){
    ListModelFilterNode* node = u_new(ListModelSortNode,1);
    node->data = data;
    node->num;
    return node;
}


#define list_model_sort_num_from_iter(ms,iter) ((iter)-(ListModelIter*)array_iter((ms)->item_array,0))


static int list_model_sort_insert_pos(ListModelSort* ms,const void* data,int start,int end){
    for(;;){
        int pos;
        int result;
        if(start >= end)
            return start;
        pos = (start+end)/2;
        result = ms->compare_func(array_get(ms->item_array,pos),data);
        if(result>0){
            end = pos;
        }else{
            start = pos+1;
        }
    }
}


static int list_model_sort_find(ListModelSort* ms,const void* data,int start,int end){
    for(;;){
        int pos;
        int result;
        if(start >= end)
            return -1;
        pos = (start+end)/2;
        result = ms->compare_func(((ListModelSortNode*)array_get(ms->item_array,pos))->data,data);
        if(result>0){
            end = pos-1;
        }else if(result<0){
            start = pos+1;
        }else{
            return pos;
        }
    }
}


static int list_model_sort_iter_compare(const void* a,const void* b,void* userdata){
    ListModelSort* ms = (ListModelSort*)userdata;
    return ms->compare_func(*(void**)a,*(void**)b);
}


static int list_model_sort_on_add(void* userdata,void* data,int num,ListModelIter* iter){
    ListModelSort* ms = (ListModelSort*)userdata;
    ListModelSortNode* node = list_model_sort_node_new(ms,num,data);
    int pos = list_model_sort_insert_pos(ms,data,0,array_size(ms->item_array));
    array_insert(ms->item_array,pos,node);
    return 0;
}


static int list_model_sort_on_remove(void* userdata,void* data,int num,ListModelIter* iter){
    ListModelSort* ms = (ListModelSort*)userdata;
    int pos = list_model_sort_find(ms,data,0,array_size(ms->item_array));
    if(pos>=0){
        array_remove(ms->item_array,pos);
    }
    return 0;
}


static void list_model_sort_done(ListModel* m){
    ListModelSort* ms = (ListModelSort*)m;
    list_model_unsubscribe_add(ms->child,ms->ss_add);
    list_model_unsubscribe_remove(ms->child,ms->ss_remove);
    list_model_unref(ms->child);
    array_free(ms->item_array);
    list_model_done_func(m);
}


static int list_model_sort_length(ListModel* m){
    ListModelSort* ms = (ListModelSort*)m;
    return array_size(ms->item_array);
}


static ListModelIter* list_model_sort_first(ListModel* m){
    ListModelSort* ms = (ListModelSort*)m;
    if(array_size(ms->item_array)<=0)
        return NULL;
    return (ListModelIter*)array_iter(ms->item_array,0);
}


static ListModelIter* list_model_sort_last(ListModel* m){
    ListModelSort* ms = (ListModelSort*)m;
    if(array_size(ms->item_array)<=0)
        return NULL;
    return (ListModelIter*)array_iter(ms->item_array,array_size(ms->item_array)-1);
}


static ListModelIter* list_model_sort_previous(ListModel* m,ListModelIter* iter){
    ListModelSort* ms = (ListModelSort*)m;
    int num = list_model_sort_num_from_iter(ms,iter) - 1;
    if(num<0)
        return NULL;
    return (ListModelIter*)array_iter(ms->item_array,num);
}


static ListModelIter* list_model_sort_next(ListModel* m,ListModelIter* iter){
    ListModelSort* ms = (ListModelSort*)m;
    int num = list_model_sort_num_from_iter(ms,iter) + 1;
    if(num>=array_size(ms->item_array))
        return NULL;
    return (ListModelIter*)array_iter(ms->item_array,num);
}


static int list_model_sort_compare(ListModel* m,ListModelIter* a,ListModelIter* b){
    if(a>b){
        return 1;
    }else if(a<b){
        return -1;
    }else{
        return 0;
    }
}


static int list_model_sort_foreach(ListModel* m,ModelCallback cb,void* userdata){
    ListModelSort* ms = (ListModelSort*)m;
    int result;
    int i;
    for(i=0;i<array_size(ms->item_array);++i){
        result = cb(userdata,array_get(ms->item_array,i),i,(ListModelIter*)array_iter(ms->item_array,i));
        if(result)
            return result;
    }
    return 0;
}


static ListModelIter* list_model_sort_get_iter(ListModel* m,int num){
    ListModelSort* ms = (ListModelSort*)m;
    return (ListModelIter*)array_iter(ms->item_array,num);
}


static int list_model_sort_get_num(ListModel* m,ListModelIter* iter){
    ListModelSort* ms = (ListModelSort*)m;
    return list_model_sort_num_from_iter(ms,iter);
}


static void* list_model_sort_get(ListModel* m,ListModelIter* iter){
    return *((void**)iter);
}


static const ListModelInterface list_model_sort_if = {
    .done     = list_model_sort_done,
    .length   = list_model_sort_length,
    .first    = list_model_sort_first,
    .last     = list_model_sort_last,
    .previous = list_model_sort_previous,
    .next     = list_model_sort_next,
    .compare  = list_model_sort_compare,
    .foreach  = list_model_sort_foreach,
    .get_iter = list_model_sort_get_iter,
    .get_num  = list_model_sort_get_num,
    .get      = list_model_sort_get
};


static int list_model_sort_new_cb(void* userdata,void* data,int num,ListModelIter* iter){
    ListModelSort* ms = (ListModelSort*)userdata;
    array_set(ms->item_array,num,data);
    return 0;
}
ListModel* list_model_sort_new(ListModel* child,ModelCompareFunc func){
    ListModelSort* ms = u_new(ListModelSort,1);
    ListModel* m = list_model_init((ListModel*)ms);
    DEBUG_ref_reg((URef*)ms,"ListModelSort");
    m->vt = &list_model_sort_if;

    ms->child = child;
    ms->item_array = array_new(list_model_length(child));
    ms->compare_func = func;
    list_model_foreach(child,list_model_sort_new_cb,ms);
    if(func)
        qsort_r(array_iter(ms->item_array,0),array_size(ms->item_array),sizeof(void*),list_model_sort_iter_compare,ms);

    ms->ss_add = list_model_subscribe_add(child,list_model_sort_on_add,ms);
    ms->ss_remove = list_model_subscribe_remove(child,list_model_sort_on_remove,ms);
    return m;
}


ListModel* list_model_sort_get_child(ListModel* m){
    return ((ListModelSort*)m)->child;
}


void list_model_sort_get_child_iter(ListModel* m,ListModelIter* src,ListModelIter* dst){
}


void list_model_sort_set_compare(ListModel* m,ModelCompareFunc func){
    ListModelSort* ms = (ListModelSort*)m;
    if(func == ms->compare_func)
        return;
    ms->compare_func = func;
    if(func)
        qsort_r(array_iter(ms->item_array,0),array_size(ms->item_array),sizeof(void*),list_model_sort_iter_compare,ms);
}
#endif
