/*
 * Copyright (C) 2016 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <fcntl.h>                      /* open() */
#include <glib.h>
#include <lauxlib.h>
#include <microhttpd.h>
#include <stdio.h>                      /* getchar() */
#include <string.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>

#include "debug.h"
#include "wgui_lua.h"


#define VERSION "0.1"


char* basepath;
int basepath_length;

#if 1
int opt_port = 8080;
static int opt_version;
static char** opt_filename;

static const GOptionEntry entries[] = {
    {"port",'p',0,G_OPTION_ARG_INT,&opt_port,"Listen to TCP port PORT","PORT"},
    {"version",'V',0,G_OPTION_ARG_NONE,&opt_version,"Display version (" VERSION ") and exit"},
    {G_OPTION_REMAINING,'\0',0,G_OPTION_ARG_FILENAME_ARRAY,&opt_filename,NULL,"<filename.lua>"},
    {NULL}
};
#endif
static lua_State* lua;


static const char* path_sep(const char* path,WguiRequest* r){
    char* p = strrchr(path,'/');
    if(!p){
        r->basepath_len = 0;
        return path;
    }
    r->basepath = path;
    r->basepath_len = p-path-1;
    return p[1]?p+1:p;
}


static struct MHD_Response* bootstrap(){
    int fd = open("bootstrap.html",O_RDONLY);
    struct stat buf;
    if(fd<0)
        return NULL;
    if(fstat(fd,&buf)<0){
        /* fstat failed */
        close(fd);
        return NULL;
    }
    return MHD_create_response_from_fd(buf.st_size,fd);
}


static struct MHD_Response* empty(){
    return MHD_create_response_from_buffer(1,"\n",0);
}


static struct MHD_Response* handle_event(WguiRequest* req,const char* event){
    char* buf;
    size_t len;

    DEBUG_p("event: %s",event);
    if('/'==*event)
        ++event;

    req->fd = open_memstream(&buf,&len);
    if(!strcmp(event,"_")){
        wgui_lua_call1s(lua,"init",req,"_");
    }else{
        wgui_lua_call0(lua,event,req);
    }
    fclose(req->fd);
    return MHD_create_response_from_buffer(len,buf,MHD_RESPMEM_MUST_FREE);
}


static int on_request(
    void* cls,
    struct MHD_Connection* conn,
    const char* url,
    const char* method,
    const char* version,
    const char* upload_data,
    size_t* upload_data_size,
    void** con_cls
){
    WguiRequest req;
    struct MHD_Response* res;
    int ret;

    DEBUG_p("URL: %s",url);
    url = path_sep(url,&req);
    DEBUG_p("URL base: %s",url);

    if(!strcmp(url,"/")){
        /* send bootstrap page */
        res = bootstrap();
    }else if(!strcmp(url,"/favicon.ico")){
        /* send favicon? */
        res = NULL;
    }else{
        /* handle event */
        res = handle_event(&req,url);
    }

    if(res){
        ret = MHD_queue_response(conn,MHD_HTTP_OK,res);
        MHD_destroy_response(res);
    }else{
        ret = MHD_queue_response(conn,MHD_HTTP_NOT_FOUND,empty());
    }

    return ret;
}


int main(int argc,char** argv){
    int filename_count;
    struct MHD_Daemon* daemon;
    {
        GOptionContext* context;
        GError* err = NULL;
        context = g_option_context_new(NULL);
        g_option_context_add_main_entries(context,entries,NULL);
        //g_option_context_set_summary(context,"");
        if(!g_option_context_parse(context,&argc,&argv,&err)){
            fprintf(stderr,"Option parsing failed: %s\n",err->message);
            g_error_free(err);
            return 1;
        }
        if(opt_version){
            printf("wgui version " VERSION "\n");
            return 0;
        }
        filename_count = opt_filename?g_strv_length(opt_filename):0;
        if(filename_count<1){
            fprintf(stderr,"%s",g_option_context_get_help(context,TRUE,NULL));
            return 1;
        }
        g_option_context_free(context);
    }

    lua = luaL_newstate();
    wgui_lua_init(lua);
    if(luaL_dofile(lua,opt_filename[0])){
        DEBUG_p("couldn't execute: %s",lua_tostring(lua,-1));
        return 1;
    }

    daemon = MHD_start_daemon(MHD_USE_SELECT_INTERNALLY,opt_port,NULL,NULL,on_request,NULL,MHD_OPTION_END);
    if(!daemon)
        return 1;
    getchar ();
    MHD_stop_daemon (daemon);

    wgui_lua_done(lua);
    lua_close(lua);

    return 0;
}
