/*
 * Copyright (C) 2016 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <stdarg.h>
#include <string.h>

#include "memchunk.h"
#include "memutil.h"


#define MINIMUM_SIZE 256


static void memchunk_clear(Memchunk* mc){
    if(mc->is_private){
        if(mc->size)
            u_free(mc->mem);
    }else{
        if(mc->notify.func)
            mc->notify.func(mc->notify.data);
        mc->is_private = 1;
    }
    mc->size = 0;
}


static size_t get_alloc_size(Memchunk* mc,size_t request_size){
    size_t alloc_size;
    if(mc->greedy){
        for(alloc_size=MINIMUM_SIZE;alloc_size<request_size;alloc_size<<=1);
    }else{
        alloc_size = request_size;
    }
    return alloc_size;
}



Memchunk* memchunk_init(Memchunk* mc,int flags){
#ifdef USE_REFCOUNT
    /* Hack to prevent non dynamic objects from being affected by unrefing. */
    mc->ref.count = 0;
#endif
    mc->size = 0;
    mc->greedy = flags&MEMCHUNK_GREEDY?1:0;
    mc->is_private = 1;
    mc->is_writable = 1;
    return mc;
}


void memchunk_done(Memchunk* mc){
    memchunk_clear(mc);
}


static void memchunk_free_internal(Memchunk* mc){
    memchunk_done(mc);
    u_free(mc);
}


Memchunk* memchunk_new(int flags){
    Memchunk* mc = memchunk_init(u_new(Memchunk,1),flags);
#ifdef USE_REFCOUNT
    mc->ref.count = 1;
    mc->ref.destroy = (URefDestroyFunc)memchunk_free_internal;
    DEBUG_ref_reg((URef*)mc,"Memchunk");
#endif
    return mc;
}


#ifndef USE_REFCOUNT
void memchunk_free(Memchunk* mc){
    memchunk_free_internal(mc);
}
#endif



char* memchunk_ensure_private(Memchunk* mc){
    size_t tmp_size;
    char* tmp_mem;
    if(mc->is_private)
        return mc->mem;
    tmp_size = mc->size;
    tmp_mem = mc->mem;
#ifdef USE_JEMALLOC
    mc->mem = (char*)mallocx(get_alloc_size(mc,tmp_size),0);
    mc->size = sallocx(mc->mem,0);
#else
    mc->size = get_alloc_size(mc,tmp_size);
    mc->mem = u_new(char,mc->size);
#endif
    memcpy(mc->mem,tmp_mem,tmp_size);
    if(mc->notify.func)
        mc->notify.func(mc->notify.data);
    mc->is_private = 1;
    mc->is_writable = 1;
    return mc->mem;
}


char* memchunk_ensure_private_size_nopreserve(Memchunk* mc,size_t request_size){
    if(mc->is_private && request_size<=mc->size)
        return mc->mem;
    if(!mc->is_private && mc->notify.func)
        mc->notify.func(mc->notify.data);
#ifdef USE_JEMALLOC
    mc->mem = (char*)mallocx(get_alloc_size(mc,request_size),0);
    mc->size = sallocx(mc->mem,0);
#else
    mc->size = get_alloc_size(mc,request_size);
    mc->mem = u_new(char,mc->size);
#endif
    mc->is_private = 1;
    mc->is_writable = 1;
    return mc->mem;
}


char* memchunk_ensure_writable(Memchunk* mc){
    if(mc->is_writable)
        return mc->mem;
    return memchunk_ensure_private(mc);
}


char* memchunk_ensure_writable_size_nopreserve(Memchunk* mc,size_t request_size){
    if(mc->is_writable && request_size<=mc->size)
        return mc->mem;
    return memchunk_ensure_private_size_nopreserve(mc,request_size);
}


int memchunk_try_ensure_size(Memchunk* mc,size_t request_size){
#ifdef USE_JEMALLOC
    size_t alloc_size;
    size_t new_size;
    if(mc->is_writable && request_size<=mc->size)
        return mc->size;
    if(!mc->is_private)
        return 0;
    alloc_size = get_alloc_size(mc,request_size);
    new_size = xallocx(mc->mem,request_size,alloc_size-request_size,0);
    if(new_size>=request_size)
        return mc->size = new_size;
#endif
    return 0;
}


char* memchunk_ensure_size_full(Memchunk* mc,size_t request_size,const char* preserve_begin,size_t preserve_size){
    off_t offs;
    /* Special case if the chunk is empty, assume preserve_begin and
       preserve_size cannot be valid. */
    if(mc->size<=0){
#ifdef USE_JEMALLOC
        mc->mem = (char*)malloc(get_alloc_size(mc,request_size));
        mc->size = sallocx(mc->mem,0);
#else
        mc->size = get_alloc_size(mc,request_size);
        mc->mem = u_new(char,mc->size);
#endif
        mc->is_private = 1;
        mc->is_writable = 1;
        if(mc->size<=0)
            return NULL;
        return mc->mem;
    }
    if(!preserve_begin)
        preserve_begin = mc->mem;
    offs = preserve_begin - mc->mem;
    if(offs+request_size<=mc->size)
        return (char*)preserve_begin;
#ifdef USE_JEMALLOC
    if(mc->is_private && memchunk_try_ensure_size(mc,offs+request_size)){
        /* The chunk could be expanded in place. */
        return (char*)preserve_begin;
    }else if(mc->is_writable && request_size<=mc->size){
        /* The chunk could be used by moving the used part. */
        memmove(mc->mem,preserve_begin,preserve_size);
        return mc->mem;
    }else{
        /* Make a new allocation. */
        char* tmp = mc->mem;
        mc->mem = (char*)mallocx(get_alloc_size(mc,request_size),0);
        mc->size = sallocx(mc->mem,0);
        if(preserve_size>0)
            memcpy(mc->mem,preserve_begin,preserve_size);
        if(mc->is_private){
            u_free(tmp);
        }else if(mc->notify.func){
            mc->notify.func(mc->notify.data);
        }
        mc->is_private = 1;
        mc->is_writable = 1;
        return mc->mem;
    }
#else
    /* FIXME!! perhaps use realloc in some corner cases? */
    if(request_size<=mc->size){
        /* The chunk could be used by moving the used part. */
        memmove(mc->mem,preserve_begin,preserve_size);
        return mc->mem;
    }else{
        /* Make a new allocation. */
        char* tmp = mc->mem;
        mc->size = get_alloc_size(mc,offs+request_size);
        mc->mem = u_new(char,mc->size);
        if(preserve_size>0)
            memcpy(mc->mem,preserve_begin,preserve_size);
        if(mc->is_private){
            u_free(tmp);
        }else if(mc->notify.func){
            mc->notify.func(mc->notify.data);
        }
        mc->is_private = 1;
        mc->is_writable = 1;
        return mc->mem;
    }
#endif
}



Memchunk* memchunk_set(Memchunk* mc,const char* mem,size_t size,int flags,...){
    if(flags&MEMCHUNK_COPY){
        memchunk_ensure_size_nopreserve(mc,size);
        memcpy(mc->mem,mem,size);
        mc->is_private = 1;
        mc->is_writable = 1;
    }else{
        mc->mem = (char*)mem;
        mc->size = size;
        if(flags&MEMCHUNK_STEAL){
            mc->is_private = 1;
            mc->is_writable = 1;
        }else{
            mc->is_private = 0;
            mc->is_writable = flags&MEMCHUNK_WRITABLE?1:0;
            if(flags&MEMCHUNK_NOTIFY){
                va_list ap;
                va_start(ap,flags);
                mc->notify.func = va_arg(ap,MemchunkNotify);
                mc->notify.data = va_arg(ap,void*);
            }else{
                mc->notify.func = NULL;
            }
        }
    }
    return mc;
}
