/*
 * Copyright (C) 2016 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include "config.h"
#include "priv.h"

#include <unistd.h>

#include "debug.h"
#include "genjs.h"
#include "memutil.h"
#include "namespace.h"
#include "session.h"
#include "strutil.h"
#include "widgets/root.h"

static HashMap* session_map;


SessionConnection* session_connection_new(
#ifdef USE_MICROHTTPD
    /* FIXME!! */
    int include_session_cookie;
#else
    HttpdConnection* conn
#endif
){
    SessionConnection* c = u_new(SessionConnection,1);
#ifdef USE_MICROHTTPD
#else
    c->conn = conn;
    c->bytes = 0;
#endif
    return c;
}


void session_connection_free(SessionConnection* c){
#ifdef USE_MICROHTTPD
#else
    httpd_connection_close(c->conn);
    c->bytes = 0;
#endif
    u_free(c);
}



static int session_callback_equal(const void* a,const void* b){
    SessionCallback* cb1 = (SessionCallback*)a;
    SessionCallback* cb2 = (SessionCallback*)b;
    return cb1->js_func == cb2->js_func && cb1->data == cb2->data;
}


static unsigned int session_callback_hash(const void* a){
    SessionCallback* cb = (SessionCallback*)a;
    return hash_combine(pointer_hash(cb->js_func),pointer_hash(cb->data));
}


Session* session_new(){
    Session* sess = u_new(Session,1);
    sess->id = u_strdupf("%X%X%X%X",(unsigned int)random(),(unsigned int)random(),(unsigned int)random(),(unsigned int)random ());
    strn_init(&sess->buffer,1);
    sess->connection_queue = queue_new();
    sess->connection = NULL;
    ns_init(&sess->jsid_ns);
    sess->jsid_map = hash_map_new(direct_hash,direct_equal,NULL,u_free);
    ns_init(&sess->callback_ns);
    sess->callback2name = hash_map_new(session_callback_hash,session_callback_equal,u_free,NULL);
    sess->name2callback = hash_map_new(str_hash,str_equal,u_free,NULL);
    sess->root = up_root_new(sess);
    return sess;
}


void session_free(Session* sess){
    u_free(sess->id);
    strn_done(&sess->buffer);
    queue_free_full(sess->connection_queue,(QueueNotifyFunc)session_connection_free);
    ns_done(&sess->jsid_ns);
    hash_map_free(sess->jsid_map);
    ns_done(&sess->callback_ns);
    hash_map_free(sess->callback2name);
    hash_map_free(sess->name2callback);
    up_widget_unref(sess->root);
    u_free(sess);
}


Session* session_get(const char* id){
    if(!session_map)
        return NULL;
    return (Session*)hash_map_get(session_map,id);
}


void session_add(Session* sess){
    if(!session_map)
        session_map = hash_map_new(str_hash,str_equal,NULL,(HashMapNotifyFunc)session_free);
    hash_map_add(session_map,sess->id,sess);
}



SessionConnection* session_open_connection(
    Session* sess,
#ifdef USE_MICROHTTPD
    /* FIXME!! */
#else
    HttpdConnection* conn,
#endif
    int include_session_cookie
){
    SessionConnection* c = session_connection_new(conn);
#ifdef USE_MICROHTTPD
    //sess->fd = c->pipe_fd[1];
    c->include_session_cookie = include_session_cookie;
#else
    httpd_header_ok(conn);
    if(include_session_cookie)
        httpd_header_cookie(conn,SESSION_COOKIE_NAME,sess->id);
    httpd_header_end(conn);
#endif
    queue_push_tail(sess->connection_queue,c);
    c = sess->connection = (SessionConnection*)sess->connection_queue->head->data;
#ifndef USE_MICROHTTPD
    /* send any pending bytes directly */
    {
        int len = sess->buffer.len;
        if(len>0){
            httpd_send(c->conn,sess->buffer.str,len);
            c->bytes += len;
            strn_reset(&sess->buffer);
        }
    }
#endif
    return c;
}


int session_close_connection(Session* sess){
/*
  if last_connection
    if got_data
      request new longread connection
      remove and finish this connection
    else if not got_data
      do nothing
  else if not last_connection
    if got_data
      remove and finish this connection
    else if not got data
      remove and finish this connection

  if last_connection
    if not got_data
      exit
    request new longread connection
  remove and finish this connection

*/
    SessionConnection* c = sess->connection;
    int ret;
    /* conn should be == sess.connection == sess.connection_queue.head */
    //DEBUG_p("conn       : %x",c);
    DEBUG_p("sess.conn  : %p",sess->connection);
    DEBUG_p("queue head : %p",sess->connection_queue->head->data);
    /* check against lower connection limit */
    if(queue_length(sess->connection_queue)<=1){
        if(session_bytes_for_connection(sess)<=0){
            /* just keep the connection around */
            return 1;
        }
        /* request a new longread connection */
        genjs_comet(sess);
    }
    queue_pop_head(sess->connection_queue);
    sess->connection = (SessionConnection*)(sess->connection_queue->head?sess->connection_queue->head->data:NULL);
#ifdef USE_MICROHTTPD
    {
        int len = sess->buffer.len;
        if(len>0){
            char* buf = strn_steal(&sess->buffer);
            struct MHD_Response* resp = MHD_create_response_from_buffer(len,buf.str,MHD_RESPMEM_MUST_FREE);
            strn_reset(&sess->buffer);
        }
    }
    if(c->include_session_cookie){
        char buffer[256];
        snprintf(buffer,sizeof(buffer),"%s=%s",SESSION_COOKIE_NAME,sess->id);
        MHD_add_response_header(resp,MHD_HTTP_HEADER_SET_COOKIE,buffer);
    }
    ret = MHD_queue_response(c->conn,MHD_HTTP_OK,resp);
#else
    ret = 1;
#endif
    session_connection_free(c);
    return ret;
}


int session_try_close_connection(Session* sess){
    if(!sess->connection)
        return 0;
    return session_close_connection(sess);
}


int session_bytes_for_connection(Session* sess){
    int bytes = sess->buffer.len;
#ifndef  USE_MICROHTTPD
    if(sess->connection)
        bytes += sess->connection->bytes;
#endif
    return bytes;
}


int session_send(Session* sess,const char* data,int len){
    if(sess->connection){
        sess->connection->bytes += len;
        return httpd_send(sess->connection->conn,data,len);
    }else{
        if(len<0)
            len = strlen(data);
        strn_appendn(&sess->buffer,data,len);
        return len;
    }
}


int session_sendf(Session* sess,const char* format,...){
    va_list ap;
    int len;
    va_start(ap,format);
    if(sess->connection){
        len = httpd_sendvf(sess->connection->conn,format,ap);
        sess->connection->bytes += len;
    }else{
        len = strn_appendvf(&sess->buffer,format,ap);
    }
    va_end(ap);
    return len;
}


int session_sendvf(Session* sess,const char* format,va_list ap){
    if(sess->connection){
        int len = httpd_sendvf(sess->connection->conn,format,ap);
        sess->connection->bytes += len;
        return len;
    }else{
        return strn_appendvf(&sess->buffer,format,ap);
    }
}



const char* session_get_jsid(Session* sess,void* p){
    return (const char*)hash_map_get(sess->jsid_map,p);
}


const char* session_add_jsid(Session* sess,void* p){
    char* name = ns_identifier(&sess->jsid_ns);
    hash_map_add(sess->jsid_map,p,name);
    return name;
}


static const char* session_register_callback(Session* sess,SessionCallbackType t,void* func,void* data){
    SessionCallback cb;
    char* name;
    cb.js_func = (SessionJsFunc)func;
    cb.data = data;
    name = (char*)hash_map_get(sess->callback2name,&cb);
    if(!name){
        SessionCallback* cbp = u_new(SessionCallback,1);
        *cbp = cb;
        cbp->type = t;
        name = ns_identifier(&sess->callback_ns);
        hash_map_add(sess->callback2name,cbp,name);
        hash_map_add(sess->name2callback,name,cbp);
    }
    return name;
}


const char* session_register_js_callback(Session* sess,SessionJsFunc func,void* data){
    return session_register_callback(sess,SC_JS,func,data);
}


const char* session_register_data_callback(Session* sess,SessionDataFunc func,void* data){
    return session_register_callback(sess,SC_DATA,func,data);
}


const char* session_register_file_callback(Session* sess,SessionFileFunc func,void* data){
    return session_register_callback(sess,SC_FILE,func,data);
}


SessionCallback* session_get_callback(Session* sess,const char* name){
    return (SessionCallback*)hash_map_get(sess->name2callback,name);
}
