/*
 * Copyright (C) 2016 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
/** @file
  A string class which goes to a great length for minimizing memory copies
  by taking advantage of several corner cases.
*/
#include <ctype.h>                      /* isspace() */
#include <stdarg.h>                     /* va_start() */
#include <stdio.h>                      /* snprintf() */
#include <string.h>                     /* strlen() */
#include <unicode/utf8.h>

#include "debug.h"
#include "memutil.h"
#include "strn.h"

/* --- private utility functions ----------------------------------------- */

#ifndef TEST_MEMCHUNK
static void strn_clear(Strn* sn){
    if(sn->allocation){
        u_free(sn->str);
        sn->allocation = 0;
    }
}
#endif


#ifndef TEST_MEMCHUNK
static void strn_ensure_allocation(Strn* sn,size_t len,int preserve){
    size_t allocation;
    if(len<sn->allocation)
        return;
    if(sn->greedy){
        for(allocation=256;allocation<len;allocation<<=1);
    }else{
        allocation = len;
    }
    if(sn->allocation){
#ifdef USE_JEMALLOC
        size_t size;
        size = xallocx(sn->str,allocation,0,0);
        if(size>=len){
            allocation = size;
        }else{
            char* tmp = sn->str;
            sn->str = mallocx(allocation,0);
            allocation = sallocx(sn->str,0);
            memcpy(sn->str,tmp,sn->len);
            dallocx(tmp,0);
        }
#else
        if(preserve){
            sn->str = u_renew(char,sn->str,allocation);
        }else{
            u_free(sn->str);
            sn->str = u_new(char,allocation);
        }
#endif
    }else{
        sn->str = u_new(char,allocation);
    }
    sn->allocation = allocation;
}
#endif


#ifndef TEST_MEMCHUNK
static void strn_ensure_alloc(Strn* sn){
    if(sn->type!=STRN_ALLOC){
        strn_clear(sn);
        allocation_init(&sn->alloc,0);
        sn->type = STRN_ALLOC;
        sn->writable = 1;
    }
}
#endif


static int format_length(const char* format,va_list ap2){
    va_list ap;
    int len;
    va_copy(ap,ap2);
    len = vsnprintf(NULL,0,format,ap);
    va_end(ap);
    return len;
}

/* --- public functions -------------------------------------------------- */

Strn* strn_init(Strn* sn,int greedy){
#ifdef USE_REFCOUNT
    /* Hack to prevent non dynamic objects to be affected by unrefing. */
    ((URef*)sn)->count = 0;
#endif
#ifdef TEST_MEMCHUNK
    memchunk_init(&sn->chunk,greedy?MEMCHUNK_GREEDY:0);
#else
    sn->allocation = 0;
    sn->greedy = greedy;
#endif
    sn->str = NULL;
    sn->len = 0;
    return sn;
}


void strn_done(Strn* sn){
#ifdef TEST_MEMCHUNK
    memchunk_done(&sn->chunk);
#else
    if(sn->allocation)
        u_free(sn->str);
#endif
}


static void strn_free_internal(Strn* sn){
    strn_done(sn);
    u_free(sn);
}


Strn* strn_new(int greedy){
    Strn* sn = strn_init(u_new(Strn,1),greedy);
#ifdef USE_REFCOUNT
    ((URef*)sn)->count = 1;
    ((URef*)sn)->destroy = (URefDestroyFunc)strn_free_internal;
    DEBUG_ref_reg((URef*)sn,"Strn");
#endif
    return sn;
}


#ifndef USE_REFCOUNT
void strn_free(Strn* sn){
    strn_free_internal(sn);
}
#endif

/* --- ensure functions --- */

void strn_ensure_private(Strn* sn){
#ifdef TEST_MEMCHUNK
    sn->str = memchunk_ensure_private(&sn->chunk);
#else
    const char* tmp;
    if(sn->len<=0 || sn->allocation>0)
        return;
    tmp = sn->str;
    strn_ensure_allocation(sn,sn->len+1,0);
    memcpy(sn->str,tmp,sn->len);
    sn->str[sn->len] = '\0';
    sn->zero_terminated = 1;
#endif
}


void strn_ensure_writable(Strn* sn){
#ifdef TEST_MEMCHUNK
    sn->str = memchunk_ensure_writable(&sn->chunk);
#else
    if(sn->writable)
        return;
    strn_ensure_private(sn);
#endif
}


size_t strn_ensure_private_length(Strn* sn,size_t len,int preserve){
#ifdef TEST_MEMCHUNK
    /* Add one byte for zero termination. */
    if(preserve){
        sn->str = memchunk_ensure_size_full(&sn->chunk,len+1,sn->str,sn->len);
        sn->str = memchunk_ensure_private(&sn->chunk);
    }else{
        sn->str = memchunk_ensure_private_size_nopreserve(&sn->chunk,len+1);
    }
    return sn->chunk.size;
#else
    strn_ensure_allocation(sn,len+1,1);
    return sn->allocation;
#endif
}


size_t strn_ensure_writable_length(Strn* sn,size_t len,int preserve){
#ifdef TEST_MEMCHUNK
    /* Add one byte for zero termination. */
    if(preserve){
        sn->str = memchunk_ensure_size_full(&sn->chunk,len+1,sn->str,sn->len);
        memchunk_ensure_writable(&sn->chunk);
    }else{
        sn->str = memchunk_ensure_writable_size_nopreserve(&sn->chunk,len+1);
    }
    return sn->chunk.size;
#else
    switch(sn->type){
    case STRN_ALLOC:
        sn->str = allocation_ensure_full(&sn->alloc,len+1,sn->str,sn->len);
        return sn->alloc.size;
    case STRN_CONST:
    case STRN_NOTIFY:
        if(sn->writable){
            size_t current_size = sn->len+(sn->zero_terminated?1:0);
            if(len<=current_size)
                return current_size;
        }
        /* fall through */
    case STRN_VOID:
        return strn_ensure_private_length(sn,len);
    }
    /* not reached */
    return 0;
#endif
}


const char* strn_ensure_zero_terminated(Strn* sn){
    if(!sn->zero_terminated){
#ifdef TEST_MEMCHUNK
        sn->str = memchunk_ensure_size_full(&sn->chunk,sn->len+1,sn->str,sn->len);
        sn->str = memchunk_ensure_writable(&sn->chunk);
        sn->str[sn->len] = '\0';
        sn->zero_terminated = 1;
#else
        if(STRN_ALLOC==sn->type){
            strn_ensure_private_length(sn,sn->len);
            sn->str[sn->len] = '\0';
            sn->zero_terminated = 1;
        }else{
            strn_ensure_private(sn);
        }
#endif
    }
    return sn->str;
}

/* --- set functions --- */

/**
  If length is < 0 then the string is zero terminated.
*/
Strn* strn_set_full(Strn* sn,const char* s,ssize_t len,int copy){
    int zero_terminated;
    if(!s){
        len = 0;
        copy = 0;
    }
    if((zero_terminated=len<0))
        len = strlen(s);
#ifdef TEST_MEMCHUNK
    if(copy){
        memchunk_set(&sn->chunk,s,len+1,MEMCHUNK_COPY);
        sn->str = sn->chunk.mem;
        sn->len = len;
        sn->str[len] = '\0';
        sn->zero_terminated = 1;
    }else{
        memchunk_set(&sn->chunk,s,len+(zero_terminated?1:0),0);
        sn->str = (char*)s;
        sn->len = len;
        sn->zero_terminated = zero_terminated;
    }
#else
    if(copy){
        strn_ensure_allocation(sn,len+1,0);
        memcpy(sn->str,s,(sn->len=len));
        sn->str[len] = '\0';
        sn->zero_terminated = 1;
    }else{
        strn_clear(sn);
        sn->str = (char*)s;
        sn->len = len;
        sn->type = STRN_CONST;
        sn->zero_terminated = zero_terminated;
    }
#endif
    return sn;
}


/**
  If length is < 0 then the string is zero terminated.
*/
Strn* strn_set_steal(Strn* sn,char* s,ssize_t len){
    int zero_terminated;
    if(!s){
        len = 0;
    }
    if((zero_terminated=len<0))
        len = strlen(s);
#ifdef TEST_MEMCHUNK
    memchunk_set(&sn->chunk,s,len+(zero_terminated?1:0),MEMCHUNK_STEAL);
#else
    if(sn->allocation)
        u_free(sn->str);
    sn->allocation = len;
#endif
    sn->str = s;
    sn->len = len;
    sn->zero_terminated = zero_terminated;
    return sn;
}


int strn_setf(Strn* sn,const char* format,...){
    va_list ap;
    int len;
    va_start(ap,format);
    len = strn_setvf(sn,format,ap);
    va_end(ap);
    return len;
}


int strn_setvf(Strn* sn,const char* format,va_list ap){
    int len = format_length(format,ap);
    strn_ensure_writable_length(sn,len,0);
    sn->zero_terminated = 1;
    return sn->len = vsnprintf(sn->str,len+1,format,ap);
}


Strn* strn_set_notify(Strn* sn,const char* s,ssize_t len,StrnNotifyFunc func,void* data){
    int zero_terminated;
#ifndef TEST_MEMCHUNK
    strn_clear(sn);
#endif
    if((zero_terminated=len<0))
        len = strlen(s);
#ifdef TEST_MEMCHUNK
    if(func){
        memchunk_set(&sn->chunk,s,len,MEMCHUNK_NOTIFY,func,data);
    }else{
        memchunk_set(&sn->chunk,s,len,0);
    }
    sn->str = (char*)s;
    sn->len = len;
#else
    sn->len = len;
    /* Let the type remain void if there is no data */
    if(len<=0)
        return sn;
    sn->str = (char*)s;
    if(func){
        sn->type = STRN_NOTIFY;
        sn->notify.func = func;
        sn->notify.data = data;
    }else{
        sn->type = STRN_CONST;
    }
#endif
    return sn;
}


Strn* strn_set_sub_strn(Strn* sn,Strn* master,off_t offs,ssize_t len){
#ifdef TEST_MEMCHUNK
    if(len<0)
        len = master->len-offs;
#ifdef USE_REFCOUNT
    /* Assume master is static if it lacks a refcount (at lest for the
       lifetime of sn). */
    if(master->chunk.is_private && ((URef*)master)->count>0){
        memchunk_set(&sn->chunk,master->chunk.mem,master->chunk.size,MEMCHUNK_NOTIFY,u_ref_unref,strn_ref(master));
    }else
#endif
    {
        memchunk_set(&sn->chunk,master->chunk.mem,master->chunk.size,0);
    }
    sn->str = master->str+offs;
    sn->len = len;
    sn->chunk.is_writable = master->chunk.is_writable;
#else
    strn_clear(sn);
    if(len<0)
        len = master->len-offs;
    switch(master->type){
    case STRN_VOID:
        break;
    case STRN_CONST:
        sn->str = master->str+offs;
        sn->len = len;
        sn->type = STRN_CONST;
        break;
    case STRN_ALLOC:
    case STRN_NOTIFY:
# ifdef USE_REFCOUNT
        sn->str = master->str+offs;
        sn->len = len;
        sn->type = STRN_NOTIFY;
        sn->notify.func = (StrnNotifyFunc)strn_unref;
        sn->notify.data = strn_ref(master);
# else
        strn_set_full(sn,master->str+offs,len,1);
# endif
        break;
    }
#endif
    return sn;
}

/* --- append functions --- */

Strn* strn_append_full(Strn* sn,const char* s,ssize_t len,int copy){
    /* shortcut for appending empty string */
    if(!s || len==0)
        return sn;
    /* shortcut for appending to empty string (= set string) */
    if(sn->len<=0)
        return strn_set_full(sn,s,len,copy);
    if(len<0)
        len = strlen(s);
    /*
      very special case then a constant string is appended with another
      constant string located right after it in memory already.
    */
    if(sn->str+sn->len==s && !copy){
        sn->len += len;
        return sn;
    }
#ifdef TEST_MEMCHUNK
    sn->str = memchunk_ensure_size_full(&sn->chunk,sn->len+len+1,sn->str,sn->len);
#else
    if(sn->allocation){
        strn_ensure_allocation(sn,sn->len+len+1,1);
    }else{
        const char* tmp = sn->str;
        strn_ensure_allocation(sn,sn->len+len+1,0);
        memcpy(sn->str,tmp,sn->len);
    }
#endif
    memcpy(sn->str+sn->len,s,len);
    sn->len += len;
    sn->str[sn->len] = '\0';
    sn->zero_terminated = 1;
    return sn;
}


Strn* strn_append_steal(Strn* sn,char* s,ssize_t len){
    /* shortcut for appending to empty string (= set string) */
    if(sn->len<=0)
        return strn_set_steal(sn,s,len);
    strn_append_full(sn,s,len,1);
    u_free(s);
    return sn;
}


int strn_appendf(Strn* sn,const char* format,...){
    va_list ap;
    int len;
    va_start(ap,format);
    len = strn_appendvf(sn,format,ap);
    va_end(ap);
    return len;
}


int strn_appendvf(Strn* sn,const char* format,va_list ap){
    if(sn->len<=0){
        return strn_setvf(sn,format,ap);
    }else{
        int len = format_length(format,ap);
        strn_ensure_writable_length(sn,sn->len+len,1);
        len = vsnprintf(sn->str+sn->len,len+1,format,ap);
        sn->len += len;
        return len;
    }
}


Strn* strn_append_char(Strn* sn,int ch){
#ifdef TEST_MEMCHUNK
    sn->str = memchunk_ensure_size_full(&sn->chunk,sn->len+U8_LENGTH(ch)+1,sn->str,sn->len);
#else
    if(sn->allocation){
        strn_ensure_allocation(sn,sn->len+U8_LENGTH(ch)+1,1);
    }else{
        const char* tmp = sn->str;
        strn_ensure_allocation(sn,sn->len+U8_LENGTH(ch)+1,0);
        memcpy(sn->str,tmp,sn->len);
    }
#endif
    U8_APPEND_UNSAFE(sn->str,sn->len,ch);
    sn->str[sn->len] = '\0';
    sn->zero_terminated = 1;
    return sn;
}

/* --- utlility functions --- */

void strn_trim(Strn* sn){
    size_t left = 0;
    if(sn->len<=0)
        return;
    while(isspace(sn->str[sn->len-1]) && sn->len>0)
        --sn->len;
    while(isspace(sn->str[left]) && left<sn->len)
        ++left;
    if(left>=sn->len){
        /* only whitespace in string */
        sn->len = 0;
        return;
    }
    if(left){
        sn->len -= left;
#ifdef TEST_MEMCHUNK
        sn->str += left;
#else
        if(sn->allocation){
            DEBUG_p("expensive trim");
            if(sn->zero_terminated){
                memmove(sn->str,sn->str+left,sn->len+1);
            }else{
                memmove(sn->str,sn->str+left,sn->len);
            }
        }else{
            sn->str += left;
        }
#endif
    }
}


char* strn_dup(Strn* sn){
    if(sn->len<=0)
        return NULL;
    return u_strndup(sn->str,sn->len);
}


char* strn_steal(Strn* sn){
    if(sn->len<=0)
        return NULL;
#ifdef TEST_MEMCHUNK
    if(sn->chunk.is_private && sn->str==sn->chunk.mem){
        sn->len = sn->chunk.size = 0;
        return sn->chunk.mem;
    }
#else
    if(sn->allocation){
        sn->len = 0;
        sn->allocation = 0;
        return sn->str;
    }
#endif
    return u_strndup(sn->str,sn->len);
}


char* strn_copy_to(Strn* sn,char* dst){
    memcpy(dst,sn->str,sn->len);
    return dst+sn->len;
}


/**
  Return a string which is garanteed to return a zero terminated string, it
  will maybe be a new copy which need to be freed.
*/
char* strn_maybe_dup(Strn* sn,int* copied){
    if((*copied = !sn->zero_terminated)){
        return u_strndup(sn->str,sn->len);
    }else{
        return sn->str;
    }
}



int strn_equal(const void* a,const void* b){
    Strn* sn1 = (Strn*)a;
    Strn* sn2 = (Strn*)b;
    if(sn1->len != sn2->len)
        return 0;
    return !strncmp(sn1->str,sn2->str,sn1->len);
}


unsigned int strn_hash(const void* a){
    /* 31 bit hash function */
    Strn* sn = (Strn*)a;
    unsigned int h;
    int i;
    if(sn->len<=0)
        return 0;
    h = sn->str[0];
    for(i=1;i<sn->len;++i)
        h = (h<<5)-h+sn->str[i];
    return h;
}
