/*
 * Copyright (C) 2016 Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#define TEST_MAINLOOP

#include "priv.h"

#include <netinet/in.h>
#include <netinet/tcp.h>
//#include <sys/socket.h>

#include <fcntl.h>                      /* open() */
#ifdef USE_MICROHTTPD
# include <microhttpd.h>
#elif defined(USE_ONION)
# include <onion/onion.h>
#endif
#include <unistd.h>                     /* close() */
#include <stdio.h>                      /* getchar() */
#include <string.h>                     /* strrchr() */
#include <sys/stat.h>

#include "debug.h"
#include "genjs.h"
#include "httpd.h"
#include "mainloop.h"
#include "memutil.h"
#include "session.h"
#include "underpants.css.h"
#include "upmain.h"


typedef struct{
#ifdef USE_MICROHTTPD
    struct MHD_Connection* conn;
#else
    HttpdConnection* conn;
#endif
    const char* basepath;
    int basepath_len;
}UpRequest;


#ifdef USE_ONION
static onion* httpd;
#elif defined(USE_MICROHTTPD)
static struct MHD_Daemon* httpd; /* shared among everything */
#else
static Httpd* httpd;
#endif
UpSessionFunc session_func; /* per application (looked up from port/path) */


static const char* path_sep(const char* path,UpRequest* r){
    char* p = strrchr(path,'/');
    if(!p){
        r->basepath_len = 0;
        return path;
    }
    r->basepath = path;
    r->basepath_len = p-path-1;
    return p[1]?p+1:p;
}


static int notfound(UpRequest* req){
#ifdef USE_MICROHTTPD
    struct MHD_Response* resp = MHD_create_response_from_buffer(1,"\n",0);
    return MHD_queue_response(req->conn,MHD_HTTP_NOT_FOUND,resp);
#else
    httpd_header_not_found(req->conn);
    httpd_header_end(req->conn);
    httpd_connection_close(req->conn);
    return 1;
#endif
}


static int serve_data(UpRequest* req,const char* data,ssize_t len,const char* mimetype){
#ifdef USE_MICROHTTPD
    struct MHD_Response* resp = MHD_create_response_from_data(buf.st_size,fd);
    if(mimetype)
        MHD_add_response_header(resp,"Content-Type",mimetype);
    return MHD_queue_response(req->conn,MHD_HTTP_OK,resp);
#else
    char len_s[16];
    if(len<0)
        len = strlen(data);
    sprintf(len_s,"%d",(int)len);
    httpd_header_ok(req->conn);
    httpd_header(req->conn,"Content-Length",len_s);
    if(mimetype)
        httpd_header(req->conn,"Content-Type",mimetype);
    httpd_header_end(req->conn);
    httpd_send(req->conn,data,len);
    return len;
#endif
}


static int serve_file(UpRequest* req,const char* filename,const char* mimetype){
#ifdef USE_MICROHTTPD
    int fd = open(filename,O_RDONLY);
    struct stat buf;
    if(fd<0)
        return notfound(req);
    if(fstat(fd,&buf)<0){
        /* fstat failed */
        close(fd);
        return notfound(req);;
    }
    {
        struct MHD_Response* resp = MHD_create_response_from_fd(buf.st_size,fd);
        if(mimetype)
            MHD_add_response_header(resp,"Content-Type",mimetype);
        return MHD_queue_response(req->conn,MHD_HTTP_OK,resp);
    }
#else
    return httpd_sendfile(req->conn,filename,mimetype)>=0;
#endif
}


#ifdef USE_MICROHTTPD
static int get_args_cb(void* userdata,enum MHD_ValueKind kind,const char* key,const char* value){
    *((const char**)userdata) = key;
    return MHD_NO;
}
static const char* get_args(UpRequest* req){
    const char* args = NULL;
    MHD_get_connection_values(conn,MHD_GET_ARGUMENT_KIND,&get_args_cb,&args);
    return args;
}
#else
static const char* get_args(UpRequest* req){
    return httpd_get_args(req->conn);
}
#endif


static Session* get_session(UpRequest* req){
#ifdef USE_MICROHTTPD
    const char* cookie = MHD_lookup_connection_value(req->conn,MHD_COOKIE_KIND,SESSION_COOKIE_NAME);
    Session* sess = session_get(cookie);
#else
    char* cookie = httpd_get_cookie(req->conn,SESSION_COOKIE_NAME);
    Session* sess = session_get(cookie);
    u_free(cookie);
#endif
    return sess;
}


static int handle_event(UpRequest* req,const char* event){
    DEBUG_p("event: %s",event);
    if('/'==*event)
        ++event;

    if(!strcmp(event,"!")){
        /* bootstrap event -> create new session */
        Session* sess = session_new();
        //ret = response_from_session(req,sess,1);
        session_open_connection(sess,req->conn,1);
        genjs_init(sess);
        if(session_func(NULL,sess->root)){
            session_add(sess);
        }else{
            session_free(sess);
        }
        genjs_resize(sess->root);
        return session_close_connection(sess);
#ifdef TEST_COMET
    }else if(!strcmp(event,"+")){
        /* longread */
        Session* sess = get_session(req);
        session_open_connection(sess,req->conn,0);
        return session_close_connection(sess);
#endif
    }else{
        Session* sess = get_session(req);
        SessionCallback* cb = session_get_callback(sess,event);
        /* FIXME!! handle the event of session not found */
        if(!cb){
            DEBUG_p("callback not found!");
            /* how to handle this, restart session? */
        }else{
            int ret;
            switch(cb->type){
            case SC_JS:
                //ret = response_from_session(req,sess,0);
                session_open_connection(sess,req->conn,0);
                cb->js_func(cb->data,sess,get_args(req));
                return session_close_connection(sess);
            case SC_DATA:
                {
                    Strn data;
                    char* mimetype = NULL;
                    //char len_s[16];
                    strn_init(&data,1);
                    cb->data_func(cb->data,&data,&mimetype);
#if 1
                    ret = serve_data(req,data.str,data.len,mimetype);
                    strn_done(&data);
                    return ret;
#else
                    sprintf(len_s,"%d",(int)len);
                    httpd_header_ok(req->conn);
                    httpd_header(req->conn,"Content-Length",len_s);
                    if(mimetype){
                        httpd_header(req->conn,"Content-Type",mimetype);
                        //u_free(mimetype);
                    }
                    httpd_header_end(req->conn);
                    httpd_send(req->conn,data,len);
                    //u_free(data);
                    return len;
#endif
                }
            case SC_FILE:
#if 1
                {
                    char* filename = NULL;
                    char* mimetype = NULL;
                    cb->file_func(cb->data,&filename,&mimetype);
                    if(filename){
                        ret = serve_file(req,filename,mimetype);
                        u_free(filename);
                        u_free_if_set(mimetype);
                        return ret;
                    }

                }
#else
                cb->func(cb->data,sess,NULL);
                if(sess->data.filename){
                    ret = serve_file(req,sess->data.filename,sess->data.mimetype);
                    u_free(sess->data.filename);
                    u_free_if_set(sess->data.mimetype);
                    return ret;
                }
#endif
            }
        }
    }
    return notfound(req);
}


static int on_request(
#ifdef USE_MICROHTTPD
    void* cls,
    struct MHD_Connection* conn,
    const char* url,
    const char* method,
    const char* version,
    const char* upload_data,
    size_t* upload_data_size,
    void** con_cls
#else
    void* userdata,
    HttpdConnection* conn,
    const char* url,
    const char* method,
    const char* version
#endif
){
    UpRequest req;
    //struct MHD_Response* resp;

    DEBUG_p("URL: %s",url);
    DEBUG_p("Method: %s",method);
    DEBUG_p("Version: %s",version);
    url = path_sep(url,&req);
    DEBUG_p("URL base: %s",url);

    req.conn = conn;

    if(!strcmp(url,"/")){
        /* send bootstrap page */
        //return server_file(&req,"underpants.html","text/html");
        return serve_data(&req,
            "<html><head>"
              "<title>UnderPants!</title>"
              "<link rel=\"stylesheet\" type=\"text/css\" href=\"-\" />"
              "<script>"
                "var p=[];"
                "function _(a){"
                  "var w=p.pop();"
                  "if(!w){"
                    "w=new Worker('_');"
                    "w.onmessage=function(e){"
                      "eval(e.data);"
                      "p.push(this);"
                    "};"
                  "}"
                  "w.postMessage(a);"
                "}"
              "</script>"
            "</head><body onload=\"_('!')\"></body></html>",
        -1,"text/html");
    }else if(!strcmp(url,"-")){
        DEBUG_p("sending css");
        //return serve_file(&req,"underpants.css","text/css");
        return serve_data(&req,(char*)underpants_css,underpants_css_size,"text/css");
    }else if(!strcmp(url,"_")){
        DEBUG_p("sending worker script");
        //return serve_file(&req,"underpants.js",NULL);
        return serve_data(&req,
            "self.onmessage=function(e){"
              "var x=new XMLHttpRequest;"
              "x.open('GET',e.data,false);"
              "x.send();"
              "self.postMessage(x.responseText);"
            "}",
        -1,"application/javascript");
    }else if(!strcmp(url,"/favicon.ico")){
        /* send favicon? */
        return notfound(&req);
    }else{
        /* handle event */
        return handle_event(&req,url);
    }
}


static int on_quit(void* userdata,int sig){
    mainloop_quit();
    return 0;
}



int up_init(int port,UpSessionFunc func){
    DEBUG_ref_init();
#ifdef TEST_MAINLOOP
    mainloop_init();
# ifdef USE_MICROHTTPD
    httpd = MHD_start_daemon(MHD_USE_THREAD_PER_CONNECTION,port,NULL,NULL,on_request,"",MHD_OPTION_END);
    if(!httpd)
        return 0;
# elif defined(USE_ONION)
    httpd = onion_new(O_THREADED);
# else
    httpd = httpd_new(port,on_request,NULL);
# endif

    mainloop_subscribe_signal(SIGHUP,on_quit,NULL);
    mainloop_subscribe_signal(SIGINT,on_quit,NULL);
    mainloop_subscribe_signal(SIGQUIT,on_quit,NULL);
    mainloop_subscribe_signal(SIGTERM,on_quit,NULL);
#else
    httpd = MHD_start_daemon(MHD_USE_SELECT_INTERNALLY,port,NULL,NULL,on_request,NULL,MHD_OPTION_END);
    if(!httpd)
        return 0;
#endif

    session_func = func;

    return 1;
}


void up_done(){
#ifdef USE_MICROHTTPD
    MHD_stop_daemon(httpd);
#elif defined(USE_ONION)
    /* FIXME!! */
#else
    httpd_free(httpd);
#endif
#ifdef TEST_MAINLOOP
    mainloop_done();
#else
#endif
    DEBUG_ref_done();
}


int up_main(){
#ifdef TEST_MAINLOOP
    return mainloop_run();
#else
    getchar();
    return 0;
#endif
}


void up_flush(UpWidget* w){
#ifdef TEST_COMET
    session_try_close_connection(w->session);
#endif
}
